from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Optional
from typing_extensions import override

from httpx import AsyncClient

from ..const import DEFAULT_TIMEOUT, DEFAULT_URL_TEMPLATE, Theme
from .base import BaseAkinator

if TYPE_CHECKING:
    from httpx import Cookies


HEADERS = {
    "accept-encoding": "gzip, deflate, br, zstd",
    "accept-language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7",
    "sec-ch-ua": '"Google Chrome";v="135", "Not-A.Brand";v="8", "Chromium";v="135"',
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": '"Windows"',
    "user-agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
        " AppleWebKit/537.36 (KHTML, like Gecko)"
        " Chrome/135.0.0.0 Safari/537.36"
    ),
}
PAGE_EXTRA_HEADERS = {
    "accept": (
        "text/html,application/xhtml+xml,application/xml;q=0.9"
        ",image/avif,image/webp,image/apng"
        ",*/*;q=0.8"
        ",application/signed-exchange;v=b3;q=0.7"
    ),
    "priority": "u=0, i",
    "sec-fetch-dest": "document",
    "sec-fetch-mode": "navigate",
    "sec-fetch-site": "same-origin",
    "sec-fetch-user": "?1",
    "upgrade-insecure-requests": "1",
}
XHR_EXTRA_HEADERS = {
    "accept": "*/*",
    "priority": "u=1, i",
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "x-requested-with": "XMLHttpRequest",
}


class HTTPXAkinator(BaseAkinator):
    def __init__(
        self,
        lang: str = "cn",
        theme: Optional[Theme] = None,
        child_mode: bool = False,
        base_url_template: str = DEFAULT_URL_TEMPLATE,
        allow_not_supported_lang: bool = False,
        allow_not_supported_theme: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
        **kwargs,
    ) -> None:
        super().__init__(
            lang=lang,
            theme=theme,
            child_mode=child_mode,
            base_url_template=base_url_template,
            allow_not_supported_lang=allow_not_supported_lang,
            allow_not_supported_theme=allow_not_supported_theme,
            timeout=timeout,
        )
        self.cookies: Optional[Cookies] = None
        self.cli_kwargs: dict[str, Any] = kwargs

    @asynccontextmanager
    async def create_client(self, *extra_headers: Optional[dict[str, str]]):
        headers = {**HEADERS}
        for extra in extra_headers:
            if extra:
                headers.update(extra)
        async with AsyncClient(
            headers=headers,
            follow_redirects=True,
            cookies=self.cookies,
            timeout=self.timeout,
            **self.cli_kwargs,
        ) as cli:
            try:
                yield cli
            finally:
                self.cookies = cli.cookies

    @override
    async def do_request(
        self,
        method: str,
        url: str,
        data: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> bytes:
        async with self.create_client(XHR_EXTRA_HEADERS, headers) as cli:
            resp = await cli.request(method, url, data=data)
            resp.raise_for_status()
            return resp.content

    @override
    async def goto_start_page(self) -> str:
        data = {"sid": self.theme.value, "cm": self.child_mode}
        async with self.create_client(PAGE_EXTRA_HEADERS) as cli:
            resp = await cli.post(f"{self.base_url}/game", data=data)
            resp.raise_for_status()
            return resp.text



from enum import Enum


class Theme(int, Enum):
    CHARACTERS = 1
    OBJECTS = 2
    ANIMALS = 14


class Answer(int, Enum):
    YES = 0
    NO = 1
    I_DONT_KNOW = 2
    PROBABLY = 3
    PROBABLY_NOT = 4


THEMES = {
    "en": [Theme.CHARACTERS, Theme.OBJECTS, Theme.ANIMALS],
    "ar": [Theme.CHARACTERS],
    "cn": [Theme.CHARACTERS],
    "de": [Theme.CHARACTERS, Theme.ANIMALS],
    "es": [Theme.CHARACTERS, Theme.ANIMALS],
    "fr": [Theme.CHARACTERS, Theme.OBJECTS, Theme.ANIMALS],
    "il": [Theme.CHARACTERS],
    "it": [Theme.CHARACTERS, Theme.ANIMALS],
    "jp": [Theme.CHARACTERS, Theme.ANIMALS],
    "kr": [Theme.CHARACTERS],
    "nl": [Theme.CHARACTERS],
    "pl": [Theme.CHARACTERS],
    "pt": [Theme.CHARACTERS],
    "ru": [Theme.CHARACTERS],
    "tr": [Theme.CHARACTERS],
    "id": [Theme.CHARACTERS],
}

DEFAULT_URL_TEMPLATE = "https://{}.akinator.com"
DEFAULT_TIMEOUT = 15



import base64
import json
import re
from typing import TYPE_CHECKING, Any, Optional, TypedDict
from typing_extensions import override

from ..const import DEFAULT_TIMEOUT, DEFAULT_URL_TEMPLATE, Theme
from .base import BaseAkinator

if TYPE_CHECKING:
    from playwright.async_api import Page

FORM_OP_JS_TEMPLATE = """
(el) => {
    el.querySelector("#cm").value = "{cm}"
    el.querySelector("#sid").value = "{sid}"
    el.submit()
}
"""
REQUEST_JS_TEMPLATE = """
async () => {
    const form = new FormData()
    for (const [k, v] of Object.entries({data})) {
        form.append(k, v)
    }
    const method = '{method}'
    const resp = await fetch('{url}', {
        method,
        body: ['GET', 'HEAD'].includes(method) ? undefined : form,
        headers: {headers},
    })
    const buf = new Uint8Array(await resp.arrayBuffer())
    const data = btoa(String.fromCharCode(...buf))
    return { status: resp.status, status_text: resp.statusText, data }
}
"""


class RequestReturn(TypedDict):
    status: int
    status_text: str
    data: str


def format_template(template: str, **kwargs) -> str:
    for k, v in kwargs.items():
        template = template.replace(f"{{{k}}}", str(v))  # noqa: UP032
    return template


class PlaywrightAkinator(BaseAkinator):
    def __init__(
        self,
        page: "Page",
        lang: str = "cn",
        theme: Optional[Theme] = None,
        child_mode: bool = False,
        base_url_template: str = DEFAULT_URL_TEMPLATE,
        allow_not_supported_lang: bool = False,
        allow_not_supported_theme: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
    ) -> None:
        super().__init__(
            lang=lang,
            theme=theme,
            child_mode=child_mode,
            base_url_template=base_url_template,
            allow_not_supported_lang=allow_not_supported_lang,
            allow_not_supported_theme=allow_not_supported_theme,
            timeout=timeout,
        )
        self.page = page

    @override
    async def do_request(
        self,
        method: str,
        url: str,
        data: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> bytes:
        js = format_template(
            REQUEST_JS_TEMPLATE,
            method=method,
            url=url,
            data=json.dumps(data or {}),
            headers=json.dumps(headers or {}),
        )
        resp: RequestReturn = await self.page.evaluate(js)
        if resp["status"] != 200:
            raise RuntimeError(
                f"Request failed: [{resp['status']}] {resp['status_text']}",
            )
        return base64.b64decode(resp["data"])

    @override
    async def goto_start_page(self) -> str:
        await self.page.goto(
            self.base_url,
            timeout=self.timeout * 1000,
            wait_until="load",
        )
        game_form_el = await self.page.wait_for_selector(
            "#formTheme",
            timeout=self.timeout * 1000,
            state="attached",
        )
        assert game_form_el
        js = format_template(
            FORM_OP_JS_TEMPLATE,
            cm=str(self.child_mode).lower(),
            sid=self.theme.value,
        )
        await game_form_el.evaluate(js)
        await self.page.wait_for_url(re.compile("game"), wait_until="load")
        return await self.page.content()



import asyncio
import sys
from contextlib import asynccontextmanager, suppress
from importlib.util import find_spec
from typing import TYPE_CHECKING, cast

from . import (
    THEMES,
    Answer,
    BaseAkinator,
    CanNotGoBackError,
    GameEndedError,
    HTTPXAkinator,
    PlaywrightAkinator,
    Theme,
    WinResp,
    __version__,
)

if TYPE_CHECKING:
    from playwright.async_api import Browser, Page, Playwright


@asynccontextmanager
async def use_akinator(**kwargs):
    if (pr := find_spec("patchright")) or find_spec("playwright"):
        if pr:
            from patchright.async_api import async_playwright
        else:
            from playwright.async_api import async_playwright

        playwright = None
        browser = None
        page = None
        try:
            playwright = cast("Playwright", await async_playwright().start())
            # must disable headless mode for bypassing
            browser = cast("Browser", await playwright.chromium.launch(headless=False))
            page = cast("Page", await browser.new_page())
            yield PlaywrightAkinator(page, **kwargs)
        finally:
            if page:
                await page.close()
            if browser:
                await browser.close()
            if playwright:
                await playwright.stop()

    else:
        yield HTTPXAkinator(**kwargs)


async def handle_input(aki: BaseAkinator, msg: str):
    if msg.isdigit():
        try:
            answer = Answer(int(msg) - 1)
        except Exception:
            print("Invalid answer")
            return False

        try:
            resp = await aki.answer(answer)
        except GameEndedError:
            print("You beat me!")
            return True

        if isinstance(resp, WinResp):
            should_continue = (
                input(
                    f"I guess: {resp.name_proposition} - {resp.description_proposition}\n"
                    f"Photo URL: {resp.photo} (From: {resp.pseudo})\n"
                    f"Continue? (y/N) ",
                ).lower()
                or "n"
            ) == "y"
            if not should_continue:
                return True

            print()
            await aki.continue_answer()
            return False

    elif msg == "b":
        try:
            await aki.back()
        except CanNotGoBackError:
            print("Cant go back any further!")
            print()

    else:
        print("Invalid answer")
        print()

    return False


async def main() -> int:
    print(f"CooAki v{__version__} | Console Demo")
    print()

    lang = (
        input(
            f"Available languages: {', '.join(THEMES)}\n"
            f'Input language (Defaults to "cn"): ',
        ).lower()
        or "cn"
    )
    if lang not in THEMES:
        print("Invalid language")
        return 1

    themes = THEMES[lang]
    if len(themes) == 1:
        theme = themes[0]
    else:
        default_theme = themes[0]
        theme = (
            input(
                f"Available themes: "
                f"{', '.join(f'{x} ({x.name.capitalize()})' for x in themes)}\n"
                f"Input theme (Defaults to {default_theme}): ",
            ).lower()
            or default_theme
        )
        try:
            theme = Theme(int(theme))
            assert theme in themes
        except Exception:
            print("Invalid theme")
            return 1

    child_mode = (input("Enable child mode? (y/N) ").lower() or "n") == "y"

    print()
    print(f"Using language: {lang}")
    print(f"Using theme: {theme.name.capitalize()}")
    print(f"Child mode {'enabled' if child_mode else 'disabled'}")
    print()

    async with use_akinator(
        lang=lang,
        theme=theme,
        child_mode=child_mode,
        timeout=15,
    ) as aki:
        await aki.start()

        while not aki.state.ended:
            answer_tip = ", ".join(
                f"{x + 1} ({x.name.capitalize().replace('_', ' ')})" for x in Answer
            )

            msg = input(
                f"{aki.state.step + 1}: {aki.state.question}\n"
                f"Answer: {answer_tip}, B (Back), Ctrl-C (Quit)\n"
                f"Input answer: ",
            ).lower()
            print()

            if await handle_input(aki, msg):
                break

    return 0


with suppress(KeyboardInterrupt):
    sys.exit(asyncio.run(main()))



from .base import (
    AnswerResp as AnswerResp,
    BaseAkinator as BaseAkinator,
    GameState as GameState,
    WinResp as WinResp,
)
from .httpx import HTTPXAkinator as HTTPXAkinator
from .playwright import PlaywrightAkinator as PlaywrightAkinator



import re
from abc import abstractmethod
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, Optional

from cookit.pyd import type_validate_json
from pydantic import BaseModel, ValidationError

from ..const import DEFAULT_TIMEOUT, DEFAULT_URL_TEMPLATE, THEMES, Answer, Theme
from ..errors import CanNotGoBackError, GameEndedError

GAME_REFERRER_PATH = "game"


@dataclass
class GameState:
    session: str
    signature: str

    question: str
    akitude: str = "defi.png"
    progression: float = 0.0
    step: int = 0
    win: bool = False
    step_last_proposition: Optional[int] = None
    ended: bool = False


class AnswerResp(BaseModel):
    akitude: str
    step: int
    progression: float
    question_id: int
    question: str
    completion: Optional[str] = None


class WinResp(BaseModel):
    completion: str
    id_proposition: int
    id_base_proposition: int
    valide_contrainte: int
    name_proposition: str
    description_proposition: str
    flag_photo: int
    photo: str
    pseudo: str
    nb_elements: int


class BaseAkinator:
    def __init__(
        self,
        lang: str = "cn",
        theme: Optional[Theme] = None,
        child_mode: bool = False,
        base_url_template: str = DEFAULT_URL_TEMPLATE,
        allow_not_supported_lang: bool = False,
        allow_not_supported_theme: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
    ) -> None:
        if (not allow_not_supported_lang) and (lang not in THEMES):
            raise ValueError(f"Unsupported language: {lang}")

        if not theme:
            theme = Theme.CHARACTERS
        if (not allow_not_supported_theme) and (
            (lang not in THEMES) or (theme not in THEMES[lang])
        ):
            raise ValueError(f"Unsupported theme: {theme}")

        self.lang: str = lang
        self.theme: Theme = theme
        self.child_mode: bool = child_mode
        self.base_url_template: str = base_url_template.rstrip("/")
        self.timeout: float = timeout

        self._state: Optional[GameState] = None

    @property
    def state(self) -> GameState:
        if not self._state:
            raise ValueError("Game not started")
        return self._state

    @property
    def base_url(self) -> str:
        return self.base_url_template.format(self.lang)

    @staticmethod
    def get_akitude_path(akitude: str) -> str:
        return f"assets/img/akitudes_670x1096/{akitude}"

    def get_akitude_url(self, akitude: str) -> str:
        return f"{self.base_url}/{self.get_akitude_path(akitude)}"

    async def get_akitude_image(self, akitude: str) -> bytes:
        return await self.do_request("GET", self.get_akitude_url(akitude))

    def make_answer_req_data(self):
        state = self.state
        return {
            "step": state.step,
            "progression": state.progression,
            "sid": self.theme.value,
            "cm": self.child_mode,
            "session": state.session,
            "signature": state.signature,
        }

    def handle_answer_resp(self, resp: AnswerResp):
        state = self.state
        if state.win:
            state.win = False
        state.step = resp.step
        state.progression = resp.progression
        state.question = resp.question
        state.akitude = resp.akitude

    def handle_win_resp(self, _: WinResp):
        state = self.state
        state.win = True
        state.step_last_proposition = state.step

    def ensure_not_end(self):
        state = self.state
        if state.ended:
            raise GameEndedError

    def ensure_not_win(self):
        state = self.state
        self.ensure_not_end()
        if state.win:
            raise RuntimeError(
                "Game already win, "
                "if you want to continue, please call `continue_answer`.",
            )

    def ensure_win(self):
        state = self.state
        self.ensure_not_end()
        if not state.win:
            raise RuntimeError("Game not win")

    def ensure_can_back(self):
        state = self.state
        self.ensure_not_win()
        if state.step <= 0:
            raise CanNotGoBackError

    def find_form_input_value(self, name: str, resp_text: str) -> str:
        input_reg = r'name="{0}"\s+id="{0}"\s+value="(?P<value>.+?)"'
        if not (m := re.search(input_reg.format(name), resp_text)):
            raise ValueError(f"Failed to find {name}")
        return m["value"]

    def find_question(self, resp_text: str) -> str:
        question_match = re.search(
            r'<div class="bubble-body"><p class="question-text" id="question-label">'
            r"(?P<question>.+?)"
            r"</p></div>",
            resp_text,
        )
        if not question_match:
            raise ValueError("Failed to find question")
        return question_match["question"]

    def make_referrer_headers(self, referrer_path: str = "") -> dict[str, str]:
        return {
            "origin": self.base_url,
            "referer": f"{self.base_url}/{referrer_path}",
        }

    @abstractmethod
    async def do_request(
        self,
        method: str,
        url: str,
        data: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> bytes: ...

    async def request(
        self,
        method: str,
        path: str,
        data: Optional[dict[str, Any]] = None,
        referrer_path: Optional[str] = "",
    ) -> bytes:
        headers = {}
        if referrer_path:
            headers.update(self.make_referrer_headers(referrer_path))
        return await self.do_request(
            method,
            f"{self.base_url}/{path}",
            data=data,
            headers=headers,
        )

    @abstractmethod
    async def goto_start_page(self) -> str: ...

    async def start(self):
        resp_text = await self.goto_start_page()
        self._state = GameState(
            self.find_form_input_value("session", resp_text),
            self.find_form_input_value("signature", resp_text),
            self.find_question(resp_text),
        )
        return self._state

    async def answer(self, answer: Answer):
        state = self.state
        self.ensure_not_win()

        data = {
            **self.make_answer_req_data(),
            "answer": answer.value,
            "step_last_proposition": (x if (x := state.step_last_proposition) else ""),
        }
        resp_text = await self.request(
            "POST",
            "answer",
            data=data,
            referrer_path="game",
        )

        with suppress(ValidationError):
            self.handle_answer_resp(resp := type_validate_json(AnswerResp, resp_text))
            return resp

        with suppress(ValidationError):
            self.handle_win_resp(resp := type_validate_json(WinResp, resp_text))
            return resp

        state.ended = True
        state.akitude = "deception.png"
        raise GameEndedError

    async def continue_answer(self):
        self.ensure_win()

        data = self.make_answer_req_data()
        resp_text = await self.request(
            "POST",
            "exclude",
            data=data,
            referrer_path="game",
        )
        self.handle_answer_resp(resp := type_validate_json(AnswerResp, resp_text))
        return resp

    async def back(self):
        self.ensure_can_back()

        data = self.make_answer_req_data()
        resp_text = await self.request(
            "POST",
            "cancel_answer",
            data=data,
            referrer_path="game",
        )
        self.handle_answer_resp(resp := type_validate_json(AnswerResp, resp_text))
        return resp



from .akinator import (
    AnswerResp as AnswerResp,
    BaseAkinator as BaseAkinator,
    GameState as GameState,
    HTTPXAkinator as HTTPXAkinator,
    PlaywrightAkinator as PlaywrightAkinator,
    WinResp as WinResp,
)
from .const import THEMES as THEMES, Answer as Answer, Theme as Theme
from .errors import (
    CanNotGoBackError as CanNotGoBackError,
    GameEndedError as GameEndedError,
)

Akinator = HTTPXAkinator

__version__ = "0.3.0.post1"
